//SOLUZIONI

import java.util.*;

public class Esame{
  
  
  private static class Brano implements Comparable<Brano> {
    public double adeguatezza; //adeguatezza del brano per l'utente  
    public String nome; //nome del brano
    public int numAscolti; //numero di volte che il brano è stato ascoltato

	/**
	* Costruttore con 3 argomenti
	*
	* @param a adeguatezza
	* @param n nome
	* @param na numero ascolti
	*/
    public Brano(String n, double a, int na){
      adeguatezza = a;
      nome = n;
      numAscolti = na;	
    }

    public String toString(){
	    return "<"+nome+", "+adeguatezza+", "+numAscolti+">";
    }

   /** 	
   *Implementa il metodo compareTo per oggetti di tipo brano. (Punti 25)
    * Un brano b1 e' minore del brano b2 se l'adeguatezza del brano b1 e' minore dell'adeguatezza del brano b2. Altrimenti b1 e' maggiore di b2.
    * Se i brani b1 e b2 hanno lo stesso valore di adeguatezza, il brano b1 e' minore (o maggiore) del brano b2 se il numero di ascolti di b1 e' minore (o maggiore) del numero di ascolti di b2. 
    * Se i brani b1 e b2 hanno stesso valore di adeguatezza e stesso numero di ascolti, il brano b1 e' minore (o maggiore) del brano b2 se il nome di b1 precede (o segue) il nome di b2 secondo l'ordine lessicografico. 
    * @param b brano con cui comparare l'oggetto di invocazione
    * @return un intero negativo/positivo/nullo se l'oggetto di invocazione e' minore/maggiore/uguale al parametro.
    */
    public int compareTo(Brano b){
		double compAdeguatezza = adeguatezza - b.adeguatezza;
		if(compAdeguatezza==0){
			int compNA = numAscolti - b.numAscolti;
			if (compNA==0){
				return nome.compareTo(b.nome);					
			} else {
				return Integer.signum(compNA); 
			}
		} else {
			return compAdeguatezza>0?1:-1;
		}
    }	
 } 

  
  /**
  * data una matrice di double, ritorna il valore del prodotto più alto per le righe. (Punti 25) 
  * Ad esempio data la matrice 
  * {{ 3, 0.5, 6 },
  *  { -8, -3, 0.5 }},
  * il metodo ritorna
  * 	  12 (-8*-3*.5= 12 > 3*.5*6=9).
  * se la matrice non contiene nessun elemento il metodo ritorna 0..
  *
  *  @param m la matrice di double in ingresso
  *  @return un array che contiene la riga che ha il prodotto massimo dei suoi elementi.
  */	
  public static double maxRowProd(double[][] m){
		boolean found = false;
		double maxProd = 0.;	
		for (int r=0; r<m.length;r++){	
			double currProd = 1.0;
			for (int c=0; c<m[0].length; c++){	
				currProd*= m[r][c];				
			}
			if (!found || (currProd > maxProd)){					
				maxProd= currProd;
				found=true;
			}
		}
	    	return maxProd;
  }
  
 
  /**
  * Data un insieme di brani, ritornare una mappa che associa a ciascun numero di ascolti il nome del brano che ha adeguatezza più alta tra quelli che hanno lo stesso numero di ascolti. (Punti 25)
  * Ad esempio se b = {("ants marching",15.0,3), ("Hanging around",25.5,7), ("ramblin' man",21.3,8),("two steps",14.5,3), ("shadow stabbing",28,7)} la mappa da restituire sara' {3="ants marching", 7="shadow stabbing", 8="ramblin' man"}.
  * Se l'insieme e' vuoto il metodo deve ritornare una mappa vuota. 	
  *
  * @param s insieme di voli
  * @return mappa che associa a ciascuna destinazione presente nella mappa il costo minore tra i voli per quella destinazione.
  */
  public static Map<Integer,Brano> bestSong(Set<Brano> b){
    	Map<Integer,Brano> res = new HashMap<Integer,Brano>(); 
	for (Iterator<Brano> bit = b.iterator(); bit.hasNext();){
		Brano br = bit.next();
		if (res.keySet().contains(br.numAscolti)) {
			Brano best = res.get(br.numAscolti);
			if (br.adeguatezza > best.adeguatezza){
				best = br;		
			} 
			res.put(br.numAscolti,best);
		} else {
			res.put(br.numAscolti,br);
		}
    }
    return res;
  }

  /**
  * Dato un grafo g, un nodo n ed un insieme di nodi s ritornare true se tutti i nodi in s sono raggiungibili dal nodo n e false altrimenti. Un nodo k è raggiungibile da un nodo n se esiste un cammino che connette k ad n.  (Punti 25).
  * Assumere che il grafo contenga sempre almeno il nodo passato come parametro.	
  *
  * @param g grafo in ingresso
  * @param n nodo di g 
  * @return mappa che associa a ciascun nodo k di g un valore true o false. True se k è raggiungibile da n e false altrimenti
  */
  public static boolean reachSet(Grafo g, Object n, Set<Object> s){
		Set<Object> marking = new HashSet<Object>();
		dfs(g,n,marking);
		return marking.containsAll(s);		
  }

  public static void dfs(Grafo g, Object node, Set<Object> nodes){
		nodes.add(node);		
		Set<Arco> edges = g.getEdgeSet(node);
		for(Arco a: edges){
			if (!nodes.contains(a.getNode1())){
				dfs(g,a.getNode1(),nodes);
			}
			if (!nodes.contains(a.getNode2())){
				dfs(g,a.getNode2(),nodes);
			}
		} 
  }	

  public static void main(String[] args){

 
    System.out.println("**********TEST equals()*************");	

	Brano b1 = new Brano("ants marching",15.0,3);
	Brano b2 = new Brano("hanging around",25.5,7);
	Brano b3 = new Brano("ramblin' man",21.3,8);
	Brano b4 = new Brano("two steps",14.5,3);
	Brano b5 = new Brano("shadow stabbing",28,7);
	Brano b6 = new Brano("mr Jones",25.5,6);
	Brano b7 = new Brano("typical situations",25.5,6);


	System.out.println("Brano "+b1+" comparato con "+b2+" "+b1.compareTo(b2)+" [dovrebbe essere negativo]");

	System.out.println("Brano "+b2+" comparato con "+b6+" "+b2.compareTo(b6)+" [dovrebbe essere positivo]");

	System.out.println("Brano "+b6+" comparato con "+b7+" "+b6.compareTo(b7)+" [dovrebbe essere negativo]");
    
	System.out.println("Brano "+b3+" comparato con "+b3+" "+b3.compareTo(b3)+" [dovrebbe essere nullo]");


    System.out.println("**********TEST maxColElem()*************");

	double[][] m1= {
       { 3, 0.5, 6 },
       { -8, -3, 0.5 }
      };

	double[][] m2= {
      { -0.5, -0.9 },
      { -0.3, -0.1 },
      { 0.9, -0.7 },
      };

	double[][] m3= new double[0][0];


	double res1 = maxRowProd(m1);
    System.out.println("massimo prodotto per riga per la matrice  "+Arrays.deepToString(m1)+" \n "+ res1 +" [dovrebbe essere 12]");

	double res2 = maxRowProd(m2);
    System.out.println("massimo per colonna per la matrice  "+Arrays.deepToString(m2)+" \n "+ res2+" [dovrebbe essere 0.45]");

	double res3 = maxRowProd(m3);
    System.out.println("massimo per colonna per la matrice  "+Arrays.deepToString(m3)+" \n "+res3+" [dovrebbe essere 0.0]");
      
    System.out.println("**********TEST cheapestDestination()*************");

	Set<Brano> s = new HashSet<Brano>();
	s.add(b1); s.add(b2); s.add(b3); s.add(b4); s.add(b5);
	 
	Map<Integer,Brano> resMap = bestSong(s);
 
	System.out.println("Insieme di brani : " + s + "\n Brano con maggior numero di ascolti per lo stesso valore di adeguatezza : "+resMap+ " dovrebbe essere {3=(ants marching,15.0,3), 7=(shadow stabbing,28,7), 8=(ramblin' man,21.3,8)}");

	Set<Brano> empty = new HashSet<Brano>();

	Map<Integer,Brano> resMap1 = bestSong(empty);
 
	System.out.println("Insieme di brani : " + empty + "\n Brano con maggior numero di ascolti per lo stesso valore di adeguatezza : "+resMap1+ " dovrebbe essere {}");
	
    System.out.println("**********TEST reachMap()*************");
 
	    Object na = "a"; Object nb = "b"; Object nc = "c"; Object nd = "d"; Object ne = "e"; Object nf = "f"; Object nh = "h"; Object nk = "k"; Object ni = "i"; Object nj = "j";

	    Grafo g1 = new Grafo("g1");    
	    g1.add(na,nb,0);
	    g1.add(nb,nc,0);
	    g1.add(nc,nd,0);
	    g1.add(nc,ne,0);
	    
	    System.out.println("L'insieme di archi per il grafo g1: " + g1.getEdgeSet());
		
           Set<Object> nodes1 = new HashSet<Object>();	   
	   nodes1.add(na); nodes1.add(nb); nodes1.add(nc); 			

 	
	   System.out.println("L' insieme di nodi " +nodes1 + " è raggiungibile per il grafo g1 dal nodo a ? " + reachSet(g1,na,nodes1) +" [dovrebbe essere true]");

	    Grafo g2 = new Grafo("g2");    
	    g2.add(na,nb,0);
	    g2.add(nb,nc,0);
	    g2.add(nc,nd,0);
	    g2.add(nc,ne,0);
	    g2.add(nh,nk,0);


	    System.out.println("L'insieme di archi per il grafo g2: " + g2.getEdgeSet());

           Set<Object> nodes2 = new HashSet<Object>();	   
	   nodes1.add(na); nodes1.add(nb); nodes1.add(nk); 			

 	
	   System.out.println("L' insieme di nodi " +nodes1 + " è raggiungibile per il grafo g1 dal nodo a ? " + reachSet(g1,na,nodes1) +" [dovrebbe essere false]");

    
  }

}
